import pandas as pd
from scipy.optimize import minimize


class ModelCalibration:
    DEFAULT_PARAMETER_LOWER_BOUND = 1e-4

    def __init__(self, options):
        self.options = options

    def calibrate_model(self, model, method=None, options=None, bounds='default'):
        if bounds == 'default':
            bounds = ((self.DEFAULT_PARAMETER_LOWER_BOUND, None),) * len(model.parameters)
        loss = self._get_loss_function(model)
        res = minimize(loss, model.parameters, bounds=bounds, method=method, options=options)
        model.parameters = res.x
        return res, model

    def _get_loss_function(self, model):
        def _loss_function(parameters):
            model.parameters = parameters
            predicted_prices = pd.Series([model.price_option_black(o) for o in self.options])
            real_prices = pd.Series([o.price for o in self.options])
            return sum((predicted_prices - real_prices) ** 2)

        return _loss_function



import pandas as pd
from py_lets_be_rational.exceptions import BelowIntrinsicException
from py_vollib.black import implied_volatility as iv


def option_list_to_pandas_dataframe(options):
    return pd.DataFrame.from_records([o.to_dict() for o in options])


def pandas_dataframe_to_option_list(data_frame: pd.DataFrame):
    return [VanillaOption(**o) for o in data_frame.to_dict(orient='record')]


class VanillaOption:
    DAYS_IN_YEAR = 365.2425

    def __init__(self, instrument: str, option_type, date, price: float, strike: float, spot: float,
                 maturity, dividend=0):
        self.instrument = instrument
        self.option_type = option_type.lower()
        self.date = date
        self.price = price
        self.strike = strike
        self.spot = spot
        self.maturity = maturity
        self.dividend = dividend

    @property
    def years_to_maturity(self) -> float:
        return (self.maturity - self.date).days / self.DAYS_IN_YEAR

    @property
    def implied_volatility_of_undiscounted_price(self):
        try:
            return iv.implied_volatility_of_undiscounted_option_price(
                self.price,
                self.spot,
                self.strike,
                self.years_to_maturity,
                self.option_type
            )
        except BelowIntrinsicException:
            return 0

    def to_dict(self):
        return {k: v for k, v in vars(self).items() if not k.startswith('_')}



import abc

import numpy as np
from scipy import linalg as la

from vanilla_option_pricing.option_pricing import OptionPricingModel


class PossiblePricingModel(abc.ABC):
    def as_option_pricing_model(self):
        return OptionPricingModel(self)


class LogMeanRevertingToGeneralisedWienerProcess(PossiblePricingModel):

    def __init__(self, p_0: np.matrix, l: float, s_x: float, s_y: float):
        self.p_0 = p_0
        self.l = l
        self.s_x = s_x
        self.s_y = s_y

    @property
    def parameters(self):
        return [self.l, self.s_x, self.s_y]

    @parameters.setter
    def parameters(self, value):
        self.l = value[0]
        self.s_x = value[1]
        self.s_y = value[2]

    def variance(self, t):
        first_term = (self.p_0[0, 0] - 2 * self.p_0[1, 0] + self.p_0[1, 1] - (self.s_x ** 2 + self.s_y ** 2) / (
                2 * self.l)) * np.exp(-2 * self.l * t)
        second_term = 2 * (self.p_0[1, 0] - self.p_0[1, 1] + self.s_y ** 2 / self.l) * np.exp(-self.l * t)
        third_term = (self.s_y ** 2) * t + (self.s_x ** 2 - 3 * (self.s_y ** 2)) / (2 * self.l) + self.p_0[1, 1]
        return first_term + second_term + third_term


class OrnsteinUhlenbeck(PossiblePricingModel):

    def __init__(self, p_0: float, l: float, s: float):
        self.p_0 = p_0
        self.l = l
        self.s = s

    @property
    def parameters(self):
        return [self.l, self.s]

    @parameters.setter
    def parameters(self, value):
        self.l = value[0]
        self.s = value[1]

    def variance(self, t):
        return self.p_0 * np.exp(-2 * self.l * t) + self.s ** 2 / (2 * self.l) * (1 - np.exp(-2 * self.l * t))


class BlackScholes(PossiblePricingModel):

    def __init__(self, s: float):
        self.s = s

    @property
    def parameters(self):
        return [self.s]

    @parameters.setter
    def parameters(self, value):
        self.s = value[0]

    def variance(self, t):
        return self.s ** 2 * t


class NumericalLogMeanRevertingToGeneralisedWienerProcess(PossiblePricingModel):

    def __init__(self, p_0: np.matrix, l: float, s_x: float, s_y: float):
        self.p_0 = p_0
        self.l = l
        self.s_x = s_x
        self.s_y = s_y
        self.numerical_model = NumericalModel(self.__get_A_matrix(), self.__get_B_matrix(), self.p_0)

    @property
    def parameters(self):
        return [self.l, self.s_x, self.s_y]

    @parameters.setter
    def parameters(self, value):
        self.l = value[0]
        self.s_x = value[1]
        self.s_y = value[2]
        self.numerical_model.parameters = [self.__get_A_matrix(), self.__get_B_matrix()]

    def variance(self, t):
        return self.numerical_model.variance(t)

    def __get_A_matrix(self):
        return np.matrix([[-self.l, self.l], [0, 0]])

    def __get_B_matrix(self):
        return np.matrix([[self.s_x, 0], [0, self.s_y]])


class NumericalModel:

    def __init__(self, A: np.matrix, B: np.matrix, p_0: np.matrix):
        self.A = A
        self.B = B
        self.p_0 = p_0

    @property
    def parameters(self):
        return [self.A, self.B]

    @parameters.setter
    def parameters(self, value):
        self.A = value[0]
        self.B = value[0]

    def variance(self, t):
        dim = self.A.shape[0]
        F = la.expm(
            np.bmat([
                [self.A, self.B * np.transpose(self.B)],
                [np.zeros_like(self.A), -np.transpose(self.A)]
            ]) * t)
        P = (F[0:dim, 0:dim] * self.p_0 + F[0:dim, dim:2 * dim]) * la.inv(F[dim:2 * dim, dim:2 * dim])
        return P[0, 0]



from setuptools import setup, find_packages

from docs.conf import version

setup(
    name='vanilla_option_pricing',
    version=version,
    packages=find_packages(),
    url='',
    license='MIT',
    author='Emanuele Fabbiani',
    author_email='emanuele.fabbiani@xtreamers.io',
    description='A very simple library to calibrate mean-reverting stochastic models and compute their variance. '
                'Includes vanilla option pricing functionalities.',
    install_requires=['pandas', 'py_vollib', 'scipy', 'numpy', 'py_lets_be_rational'],
    test_suite='nose.collector',
    tests_require=['nose']
)



import numpy as np
from py_vollib.black import undiscounted_black
from py_vollib.black_scholes_merton import black_scholes_merton


class OptionPricingModel:

    def __init__(self, model):
        self.model = model

    @property
    def parameters(self):
        return self.model.parameters

    @parameters.setter
    def parameters(self, value):
        self.model.parameters = value

    def variance(self, t):
        return self.model.variance(t)

    def standard_deviation(self, t):
        return np.sqrt(self.variance(t))

    def volatility(self, t):
        return self.standard_deviation(t) / np.sqrt(t)

    def price_black_scholes_merton(self, option_type, spot, strike, years_to_maturity, risk_free_rate, dividend=0):
        volatility = self.volatility(years_to_maturity)
        price = black_scholes_merton(option_type, spot, strike, years_to_maturity, risk_free_rate, volatility, dividend)
        return price

    def price_option_black_scholes_merton(self, option, risk_free_rate: float):
        return self.price_black_scholes_merton(
            option.option_type,
            option.spot,
            option.strike,
            option.years_to_maturity,
            risk_free_rate,
            option.dividend
        )

    def price_black(self, option_type, spot, strike, years_to_maturity):
        volatility = self.volatility(years_to_maturity)
        price = undiscounted_black(spot, strike, volatility, years_to_maturity, option_type)
        return price

    def price_option_black(self, option):
        return self.price_black(option.option_type, option.spot, option.strike, option.years_to_maturity)
